# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

import wx
import wx.xrc
import wx.glcanvas
import wx.lib.delayedresult
import math
import time
import os, os.path

from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
import PGOpenGLUtils

import GameConfig
import GameUIController
import PGPieceAnimation
import Preferences
import WXUtils, PGUtils
import PGAppInfo

sin = math.sin
cos = math.cos

LIFT_MAPPINGS = {'Q': (0,1), 'A': (0,-1), 'W': (1,1), 'S': (1,-1), 'E': (2,1), 'D': (2,-1)}
LIFT_VISIBILITY_MAPPINGS = {'Z': 0, 'X': 1, 'C': 2}

RADIUS_2D = 1.75
LIFT_DELTA = 0.3
ANIMATION_INTERVAL = 25

class PGOpenGLBoardView (wx.glcanvas.GLCanvas):
    def __init__(self, parent):
        wx.glcanvas.GLCanvas.__init__(self, parent, -1)
        self.init = False
        self.Bind(wx.EVT_PAINT, self.paint)
        self.Bind(wx.EVT_SIZE, self.resize)
        
        self.Bind(wx.EVT_KEY_DOWN, self.keydown)
        
        self.Bind(wx.EVT_RIGHT_DOWN, self.start_drag)
        self.Bind(wx.EVT_RIGHT_UP, self.stop_drag)
        self.Bind(wx.EVT_MOTION, self.motion_event)
        self.Bind(wx.EVT_LEFT_DOWN, self.mouse_down)
        
        self.game_state = None
        
        self.reset_animations()
        self.reset_view()

    def reset_animations(self):
        self.active_animations = None
        self.animation_to_state = None
        self.animation_start_time = None
        self.animation_duration = 0
        self.animation_callbacks = []
        
    def reset_view(self):
        self.theta = self.rho = 0.0
        self.radius = 2.5
        self.lift = [0,0,0]
        self.lift_visible = [0,0,0]
        self.in_drag = False
        self.last_mousepos = None
        self.show_grid = True
        
    def player_color(self, pnum):
        return Preferences.prefs.player_color(pnum, as_float=True)
        
    def start_drag(self, event):
        self.in_drag = True
        self.dragpos = event.GetPosition()
        
    def motion_event(self, event):
        self.mouse_moved(event.GetPosition())
        
    def mouse_moved(self, pos):
        if self.in_drag:
            dx = pos[0]-self.dragpos[0]
            dy = pos[1]-self.dragpos[1]
            self.rotate(-dx/100.0, dy/100.0)
            self.dragpos = pos
            self.draw()
        else:
            if self.controller.isHumanTurn():
                p = self.cell_at_mouse_position(pos)
                self.controller.highlightCell(p)
        
    def stop_drag(self, event):
        self.in_drag = False
        
    def mouse_down(self, event):
        #print 'click at:',event.GetPosition()
        if self.controller.isHumanTurn() and not self.is_animating():
            p = self.cell_at_mouse_position(event.GetPosition())
            if p:
                if self.controller.highlightCell(p):
                    self.controller.selectHighlightedCell()
                else:
                    self.controller.undoSelection()
                    
    def periodic_timer(self):
        #print 'tick'
        spos = wx.GetMousePosition()
        wpos = self.ScreenToClient(spos)
        if wpos!=self.last_mousepos:
            self.mouse_moved(wpos)
            #print 'hittest'
        self.last_mousepos = wpos
        wx.CallLater(200, self.periodic_timer)
                

    def paint(self, event):
        wx.PaintDC(self) # needed for Windows
        self.SetCurrent()
        if not self.init:
            self.initGL()
            self.init = True
            # hack: EVT_MOTION events aren't reliable
            wx.CallLater(200, self.periodic_timer)
            
        self.draw()
        
    def resize(self, event):
        size = self.size = self.GetClientSize()
        if self.GetContext():
            self.SetCurrent()
            glViewport(0, 0, size.width, size.height)
        event.Skip()
        
    def keydown(self, event):
        code = event.GetKeyCode()
        #print 'keydown:',code
        if code==wx.WXK_UP:
            self.controller.highlightNextCell(+1)
        elif code==wx.WXK_DOWN:
            self.controller.highlightNextCell(-1)
        elif code==wx.WXK_LEFT:
            self.controller.undoSelection()
        elif code==wx.WXK_RIGHT:
            self.controller.selectHighlightedCell()
        elif 0<=code<256:
            ch = chr(code)
            if ch in LIFT_MAPPINGS:
                self.update_lift(*(LIFT_MAPPINGS[ch]))
            elif ch=='G':
                self.show_grid = not self.show_grid
                self.draw()
            elif ch=='R':
                self.reset_view()
                self.draw()
            elif ch in LIFT_VISIBILITY_MAPPINGS:
                dim = LIFT_VISIBILITY_MAPPINGS[ch]
                self.lift_visible[dim] = (1+self.lift_visible[dim]) % 3
                #print self.lift_visible
                self.draw()
            
    def update_lift(self, dim, amount):
        gdim = self.game_state.dimensions()
        if len(gdim)<3: return
        
        self.lift[dim] += amount
        if self.lift[dim]<0: self.lift[dim] = 0
        if self.lift[dim]>gdim[dim]: self.lift[dim] = gdim[dim]
        self.draw()
        
        
    def initGL(self):
        import sys
        glutInit(sys.argv)
        # set viewing projection
        light_diffuse = [1.0, 1.0, 1.0, 1.0]
        
        lights = ((GL_LIGHT0, [1, 1, 1, 0]), (GL_LIGHT1, [-1,-1,-1,0]), (GL_LIGHT2, [1, -1, 1, 1]), 
                  (GL_LIGHT3, [-1, 1, -1, -1]), (GL_LIGHT4, [1, 1, -1, 1]), (GL_LIGHT5, [-1, -1, 1, 1]) )
        for light in lights:
            glLightfv(light[0], GL_DIFFUSE, light_diffuse)
            glLightfv(light[0], GL_POSITION, light[1])

        glEnable(GL_LIGHTING)
        for light in lights:
            glEnable(light[0])
        glEnable(GL_DEPTH_TEST)
        glClearColor(0.0, 0.0, 0.0, 1.0)
        glClearDepth(1.0)

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(45.0, 1.0, 0.25, 10.0)

        glMatrixMode(GL_MODELVIEW)
    
    def setupCamera(self):
        glLoadIdentity()
        r = self.radius
        if len(self.game_state.dimensions())<3:
            r = RADIUS_2D
        gluLookAt(r*sin(self.theta)*cos(self.rho), r*sin(self.rho), r*cos(self.theta)*cos(self.rho),
                  0.0, 0.0, 0.0,
                  0.0, 1.0, 0.0)
                  
    def draw2dBoard(self, gstate):
        maxd = max(gstate.grid.shape)
        colors = Preferences.prefs.get_all_board_colors(as_float=True)
        square_half_size = 1.0/(2*maxd-2)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        for x in xrange(maxd):
            for y in xrange(maxd):
                glPushMatrix()
                glMaterialfv(GL_FRONT,GL_DIFFUSE,colors[(x+y)%2])
                
                center = self.cell_gl_translation((x,y), gstate.grid.shape)
                glBegin(GL_QUADS)
                glNormal3f( 0.0, 0.0, 1.0)
                glVertex3f(center[0]+square_half_size, center[1]+square_half_size, 0.0)
                glVertex3f(center[0]-square_half_size, center[1]+square_half_size, 0.0)
                glVertex3f(center[0]-square_half_size, center[1]-square_half_size, 0.0)
                glVertex3f(center[0]+square_half_size, center[1]-square_half_size, 0.0)
                glEnd()
                glPopMatrix()
        glDisable(GL_BLEND)
        
    def cell_gl_translation(self, coord, dim):
        def glcoord(p):
            return 1.0*p/(max(dim)-1) - 0.5
        
        x,y = glcoord(coord[0]), glcoord(coord[1])
        z = 0
        if len(dim)>2:
            z = glcoord(coord[2])
            if coord[0]<self.lift[0]: x-=LIFT_DELTA
            if coord[1]<self.lift[1]: y-=LIFT_DELTA
            if coord[2]<self.lift[2]: z-=LIFT_DELTA
        return (x,y,z)

        
    def is_coord_visible(self, c):
        if len(c)!=3: return True
        for dnum,cval in enumerate(c):
            if self.lift_visible[dnum]==1 and cval<self.lift[dnum]:  return False
            if self.lift_visible[dnum]==2 and cval>=self.lift[dnum]: return False
        return True
        
    def cell_border_size(self, state):
        ratio = 1.0 if state.grid.ndim==2 else 0.65
        return ratio/(max(state.grid.shape)-1)
        
    def iterate_positions_by_view_order(self, state, front_first=True):
        # based on rotation of board, returns iterable of positions with closest to the camera first
        if state.grid.ndim>3: return state.itercoords()
        x_range = xrange(state.grid.shape[0]) if self.theta<=0 else xrange(state.grid.shape[0]-1, -1, -1)
        y_range = xrange(state.grid.shape[1]) if self.rho<=0   else xrange(state.grid.shape[1]-1, -1, -1)
        if state.grid.ndim==3:
            z_range = xrange(state.grid.shape[2]-1, -1, -1) if -math.pi<self.theta<math.pi else xrange(state.grid.shape[2])
            sequence = ((x,y,z) for x in x_range for y in y_range for z in z_range)
        else:
            sequence = ((x,y) for x in x_range for y in y_range)
        return sequence if front_first else reversed(list(sequence))
        
    def drawGridCells(self, state):
        # if animating, use "from" state as baseline for grid values
        animation_time = 0.0
        if self.is_animating():
            animation_time = time.time()-self.animation_start_time
            
        grid = state.grid
            
        maxd = max(grid.shape)
        basescale = 1/(2.0*maxd)
        if len(state.dimensions())>2: 
            basescale /= 2.0
        
        # look for custom drawing function in [game]_UI_WX module
        try:
            ui_module = __import__(self.controller.game_config.game_name+'_UI_WX')
            drawfunc = getattr(ui_module, 'drawCell', None)
            postdrawfunc = getattr(ui_module, 'postDrawCell', None)
        except:
            drawfunc = postdrawfunc = None
                        
            
        # draw current and new pieces, using helper function
        def draw_piece(pos, val, color, scale):
            x,y,z = self.cell_gl_translation(pos, grid.shape)
            glPushMatrix()
            glTranslatef(x,y,z)
            # use custom drawing function if available
            if drawfunc is not None and drawfunc(state, pos, val, scale*basescale):
                pass
            else:
                # default: draw sphere for value of player number and "wall" for -1
                if val>0:
                    PGOpenGLUtils.draw_cylinder_or_sphere(pos, scale=scale*basescale, color=color)
                if val==-1:
                    PGOpenGLUtils.draw_wall(scale*basescale, flat=(len(pos)==2))
            glPopMatrix()
            
                
        for c in self.iterate_positions_by_view_order(state, front_first=False):
            if not self.is_coord_visible(c): continue
            animstate = self.animation_values_for_position(c, animation_time)
            draw_piece(*animstate)
            
        for animstate in self.animation_values_for_new_pieces(animation_time):
            draw_piece(*animstate)
            
        # draw grid and highlighted cells
        cubesize = self.cell_border_size(state)
        for c in self.iterate_positions_by_view_order(state):
            if not self.is_coord_visible(c): continue                    
            x,y,z = self.cell_gl_translation(c, grid.shape)
            glPushMatrix()
            glTranslatef(x,y,z)
            
            if postdrawfunc:
                postdrawfunc(state, c, cubesize/2)
            
            cubecolor = None
            if c in self.controller.highlightedCells() and not state.is_game_over():
                cubecolor = [1,1,1]
                if grid.ndim==2:
                    # put highlight cube slightly above the board, to avoid flickering in Windows
                    glTranslatef(0, 0, cubesize/4 + cubesize/1000)
                    glScalef(1, 1, 0.5)
                PGOpenGLUtils.set_color_opacity(cubecolor, 0.6)
                PGOpenGLUtils.do_with_blend(lambda: glutSolidCube(cubesize), True)
            elif grid.ndim>2 and self.show_grid:
                cubecolor = [0.3,0.3,0.3,0.3]
                glPushMatrix()
                glMaterialfv(GL_FRONT, GL_DIFFUSE, cubecolor)
                PGOpenGLUtils.do_with_blend(lambda: glutWireCube(cubesize), blend=True)
                glPopMatrix()
                
            glPopMatrix()
            
        
    def draw(self):
        state = self.game_state
        if not state: return
        self.setupCamera()
        
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        
        # the 2d board can be transparent, so draw it first if we're in front and last if we're behind
        facing_front = -math.pi/2 < self.theta < math.pi/2
        if len(state.grid.shape)==2 and facing_front:
            self.draw2dBoard(state)
            
        self.drawGridCells(self.game_state)

        if len(state.grid.shape)==2 and not facing_front:
            self.draw2dBoard(state)
            
        self.SwapBuffers()
        
    def cell_at_mouse_position(self, pos):
        state = self.game_state
        dim = state.grid.shape

        # draw cubes in back buffer with distinct colors, then figure out which one is under mouse
        glDisable(GL_DITHER)
        glDisable(GL_LIGHTING)
        self.setupCamera()
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        
        # board is packed tight in 2d, in 3d there's a gap between cells
        cubesize = self.cell_border_size(state)

        for c in state.itercoords():
            if not self.is_coord_visible(c):
                continue
            r = 8*(1+c[0])
            g = 8*(1+c[1])
            b = 8*(1+c[2]) if len(dim)>2 else 0
            glPushMatrix()
            
            glColor3ub(r, g, b)
            x,y,z = self.cell_gl_translation(c, dim)
            glTranslatef(x, y, z)
            # use "flat" cubes in 2d
            if len(dim)==2:
                glScalef(1, 1, 0.01)
            glutSolidCube(cubesize)
            
            glPopMatrix()
            
        glEnable(GL_DITHER)
        glEnable(GL_LIGHTING)
        
        # flip y coordinate and get color under pixels
        v = glGetIntegerv(GL_VIEWPORT)
        pix = glReadPixels(pos[0], v[3]-pos[1], 1, 1, GL_RGB,GL_UNSIGNED_BYTE)
        # depending on the PyOpenGL version and whether or not numpy is available, the result of glReadPixels
        # can be a (double-nested) array like [[[32 48 24]]], or a string like chr(32)+chr(48)+chr(24)
        # print 'pix=',type(pix),pix
        if isinstance(pix, basestring):
            pix = [ord(x) for x in pix]
        else:
            pix = pix[0][0]
        if pix[0]==0: return None
        c = tuple([(pix[i]//8)-1 for i in range(len(dim))])
        self.draw()
        return c
        
    def rotate(self, dx, dy):
        self.theta += dx
        self.rho += dy
        # clamp rho to +/-pi/2, keep theta between +/-pi
        if self.rho >  math.pi/2-0.01: self.rho =  math.pi/2-0.01
        if self.rho < -math.pi/2+0.01: self.rho = -math.pi/2+0.01
        if self.theta < -math.pi: self.theta += 2*math.pi
        if self.theta >  math.pi: self.theta -= 2*math.pi
        self.draw()
        #print self.theta, self.rho
        
    # animation support
    def run_animations(self, animations, newstate):
        self.animation_to_state = newstate
        if animations and animations.duration()>0:
            self.active_animations = animations
            self.animation_start_time = time.time()
            self.animation_duration = animations.duration()
            wx.CallLater(ANIMATION_INTERVAL, self.animation_update)
        else:
            self.animation_finished()
    
    def is_animating(self):
        return bool(self.active_animations)
        
    def animation_fraction_done(self):
        return min(1.0, (time.time()-self.animation_start_time)/self.animation_duration)
        
    def animation_finished(self):
        self.game_state = self.animation_to_state
        self.active_animations = None
        for callback in self.animation_callbacks:
            callback()
        self.animation_callbacks = []
        
    def animation_update(self):
        frac = self.animation_fraction_done()
        #print 'animation fraction:',frac
        if frac>=1.0:
            self.animation_finished()
        self.draw()
        if frac<1.0:
            wx.CallLater(ANIMATION_INTERVAL, self.animation_update)
            
    def blended_color_for_values(self, from_value, to_value, fraction):
        c1 = self.player_color(from_value)
        c2 = self.player_color(to_value)
        if c1 and c2:
            import PGPieceAnimation
            return PGPieceAnimation.blended_array(c1, c2, fraction)
        return None

    def animation_values_for_position(self, position, t):
        """ Returns (position, value, color, scale) tuple. value is the cell value to draw, color is the color
            (which could be a combination of old and new), scale is the size (if growing or shrinking)
        """
        val = self.game_state.grid[position]
        if not self.is_animating():
            return (position, val, self.player_color(val), 1.0)
        
        position, (fromval, toval, fraction), scale = self.active_animations.animation_state_at_time(position, val, t)
        return position, toval, self.blended_color_for_values(fromval, toval, fraction), scale
        
    def animation_values_for_new_pieces(self, t):
        if self.is_animating():
            #print 'looking for new animations'
            for position, (fromval, toval, fraction), scale in self.active_animations.new_states_at_time(t): 
                yield position, toval, self.blended_color_for_values(fromval, toval, fraction), scale
                
        
class PGApp (wx.App):
    def OnInit(self):
        # bind preferences to wx.Config to read and write to preferences file or registry
        self.SetAppName(PGAppInfo.name)
        Preferences.prefs.storage = WXUtils.ConfigDict()
        
        self.res = wx.xrc.XmlResource(os.path.join('xrc', 'pygrid.xrc'))
        self.frame = self.res.LoadFrame(None, 'mainFrame')
        
        self.menuitems = WXUtils.create_menubar(self.frame, 
            [ ('&Game', 
                [ ('newgame_menuitem', 'New Game\tCtrl+N', self.do_new_game), '-',
                  ('undo_menuitem', 'Undo Move\tCtrl+Z', self.do_undo_move),
                  ('redo_menuitem', 'Redo Move\tCtrl+Shift+Z', self.do_redo_move), '-',
                  ('open_menuitem', 'Open Game...\tCtrl+O', self.do_open_game), 
                  ('save_menuitem', 'Save Game...\tCtrl+S', self.do_save_game), '-',
                  (wx.ID_PREFERENCES, 'Preferences\tCtrl+,', self.do_prefs), '-',
                  ('host_menuitem', 'Host Game', self.do_host_network_game),
                  ('join_menuitem', 'Join Game', self.do_join_network_game), '-',
                  (wx.ID_EXIT, 'Exit\tCtrl+Q', self.do_window_closed),
                ]
              ),
              ('&Window',
                [ ('chat_menuitem', 'Show Chat Window', self.do_show_chat),
                ]
              ),
              ('&Help',
                [ (wx.ID_ABOUT, 'About Gridlock', self.do_about),
                  (wx.ID_HELP, 'Gridlock Help\tCtrl+?', self.do_main_help),
                  ('gamehelp_menuitem', 'Game Help', self.do_game_help), '-',
                  ('systeminfo_menuitem', 'System Information...', self.do_system_info), '-',
                  ('checkupdate_menuitem', 'Check for Update...', self.do_manual_update_check),
                ]
              ),
            ], )
             
        bindings = (('variationChoice', self.do_variation_choice),
                    ('pass_button', lambda event: self.controller.make_human_pass()),
                    ('undo_button', self.do_undo_move),
                    ('redo_button', self.do_redo_move),
                    ('newgame_button', self.do_new_game),
                    ('gameStatusField', ''),
                    ('gameChoice', self.doGameChoice),
                    (('playerLabels', ('player1Label', 'player2Label')), ''),
                    (('playerScoreFields', ('player1Score', 'player2Score')), ''),
                    (('playerPopups', ('player1Choice', 'player2Choice')), self.doPlayerPopupChoice),
                    )
        self.controls = WXUtils.attach_control_bindings(bindings, self.frame, self.frame.GetMenuBar())
        
        panel = wx.xrc.XRCCTRL(self.frame, 'mainPanel')
        
        # insert OpenGL view manually since it can't(?) be done in XRC
        openGLPlaceholder = wx.xrc.XRCCTRL(self.frame, 'openGLPlaceholder')
        sizer = openGLPlaceholder.GetContainingSizer()
        self.openGLView = PGOpenGLBoardView(panel)
        sizer.Remove(openGLPlaceholder)
        sizer.Add(self.openGLView, 1, wx.EXPAND)
        openGLPlaceholder.Destroy()
        self.openGLView.controller = self
        self.openGLView.SetFocus()
        
        self.frame.Bind(wx.EVT_CLOSE, self.do_window_closed)
        
        self.host_window_controller = None
        self.join_window_controller = None
        self.prefs_window_controller = None
        self.chat_window_controller = None
        self.controller = GameUIController.Controller(self)
        
        gname = Preferences.prefs.get_initial_game()
        self.controller.game_config.select_game(gname, variation=Preferences.prefs.get_variation_for_game(gname))
        self.openGLView.game_state = self.current_game_state()
        assert self.current_game_state() is not None
        
        Preferences.prefs.add_listener(self.update_ui)
        
        # popuplate game menu and initialize variations
        for g in self.game_config.game_items():
            self.controls['gameChoice'].Append(g[0], g[1])
                
        self.set_default_players()
        self.updateGameVariationPopups()
        self.updatePlayerPopups()

        panel.GetSizer().Fit(self.frame)
        # make board view square
        # Mac: need to resize frame to get correct view sizes
        # Linux: need to show window first to get correct view sizes
        self.frame.SetSize((self.frame.GetSize()[0]+1, self.frame.GetSize()[1]))
        if wx.PlatformInfo[1]=='wxGTK':
            self.frame.Show(True)
        self.controller.start_game()
        self.update_ui()
        
        diff = self.openGLView.GetSize()[0] - self.openGLView.GetSize()[1]
        if diff>0:
            self.frame.SetSize((self.frame.GetSize()[0], self.frame.GetSize()[1] + diff))
        if diff<0:
            self.frame.SetSize((self.frame.GetSize()[0] - diff, self.frame.GetSize()[1])) 
        # don't allow resizing smaller than default
        self.frame.SetSizeHints(*self.frame.GetSize())
        
        initial_rect = Preferences.prefs.get_window_rect()
        if initial_rect:
            self.frame.SetRect(initial_rect)
        
        self.frame.Show(True)
        
        if Preferences.prefs.check_for_update():
            wx.CallLater(500, self.check_for_update)
        return True
        
    game_config = property(lambda self: self.controller.game_config, None)
    
    def current_game_state(self):
        return self.controller.current_game_state()
        
    def game(self):
        return self.game_config.game
        
    def num_players(self):
        return self.current_game_state().num_players
        
    def do_about(self, event):
        info = wx.AboutDialogInfo()
        info.Name = PGAppInfo.name
        info.Version = PGAppInfo.version
        info.Copyright = PGAppInfo.copyright
        info.Description = PGAppInfo.description
        wx.AboutBox(info)
        
    def confirm_new_game(self):
        # if game is in progress, confirm restart with a dialog, if not return True directly
        if self.game().has_started() and not self.controller.is_game_over():
            dlg = wx.MessageDialog(self.frame, "The current game is in progress. Are you sure you want to restart?",
                                   "Restart", wx.YES_NO | wx.ICON_INFORMATION)
            return (dlg.ShowModal()==wx.ID_YES)
        return True
        
    def start_new_game(self):
        self.openGLView.reset_animations()
        self.controller.start_game()
        
    def do_new_game(self, event):
        if self.confirm_new_game():
            self.game_config.start_game()
            self.start_new_game()
            
    def set_default_players(self):
        np = self.num_players()
        if np>1 and not self.game_config.network_players():
            ais = [Preferences.prefs.get_ai_for_game(self.game_config.game_name, pnum+1) for pnum in range(np)]
            if all(a is None for a in ais):
                # no prefs, default 2nd (last) opponent to first AI (should be cleaner to access)
                ais[-1] = self.game_config.ai_items()[0][1]
                Preferences.prefs.set_ai_for_game(self.game_config.game_name, np, ais[-1])
                
            for pnum, ai in enumerate(ais):
                self.game_config.set_ai_for_player(ai, pnum+1)

            # don't start with all players AI-controlled
            if all(self.game_config.player_type(pnum+1)==GameConfig.AI_PLAYER for pnum in range(np)):
                self.game_config.set_human_player(1)
                Preferences.prefs.set_ai_for_game(self.game_config.game_name, 1, GameConfig.HUMAN_PLAYER)            
        
    def do_host_network_game(self, event):
        import PGHostWindowController
        if not self.host_window_controller:
            self.host_window_controller = PGHostWindowController.Controller(self.controller)
        self.host_window_controller.showWindow()
        
    def do_join_network_game(self, event):
        import PGJoinWindowController
        if not self.join_window_controller:
            self.join_window_controller = PGJoinWindowController.Controller(self.controller)
        self.join_window_controller.showWindow()
        
    def do_prefs(self, event):
        import PGPrefsWindowController
        if not self.prefs_window_controller:
            self.prefs_window_controller = PGPrefsWindowController.Controller()
        self.prefs_window_controller.showWindow()
        
    def isHumanTurn(self):
        return self.controller.is_human_turn()
        
    def do_undo_move(self, event):
        self.controller.undo_move()
        
    def do_redo_move(self, event):
        self.controller.redo_move()
        
    def do_save_game(self, event):
        dialog = wx.FileDialog(self.frame, defaultFile="", style=wx.SAVE)
        if dialog.ShowModal()==wx.ID_OK:
            path = dialog.GetPath()
            self.controller.save_game(path)
        
    def do_open_game(self, event):
        dialog = wx.FileDialog(self.frame, defaultFile="", style=wx.OPEN | wx.CHANGE_DIR)
        if dialog.ShowModal()==wx.ID_OK:
            path = dialog.GetPath()
            self.controller.load_game(path)
            self.openGLView.game_state = self.current_game_state()
            self.updateGameVariationPopups()
            self.updatePlayerPopups()
            self.update_ui()
            
    def do_main_help(self, event):
        PGUtils.open_url_from_path(os.path.join('help', 'index.html'))
    
    def do_game_help(self, event):
        PGUtils.open_url_from_path(os.path.join('help', 'games', self.game_config.game_name+'.html'))
        
    def do_system_info(self, event):
        """ show system, Python, and library version info, for debugging """
        import platform, OpenGL, GridUtils, wx.lib.dialogs
        j = PGUtils.import_json()
        msg = '\n'.join([PGAppInfo.name+': '+PGAppInfo.version,
               'Platform: '+platform.platform(True, False),
               'Python: '+platform.python_version(),
               '\t'+', '.join(platform.python_build()),
               '\t'+platform.python_compiler(),
               'wxPython: '+wx.version(),
               'PyOpenGL: '+OpenGL.__version__,
               'numpy: '+GridUtils.import_numpy().version.version,
               'json: '+j.__name__+' '+j.__version__
        ])
        dlg = wx.lib.dialogs.ScrolledMessageDialog(self.frame, msg, "System Information")
        dlg.ShowModal()
        
    def do_show_chat(self, event):
        if not self.chat_window_controller:
            import PGChatWindowController
            self.chat_window_controller = PGChatWindowController.Controller(self.controller)
        self.chat_window_controller.show_window()    
        
    # game controller delegate methods start
    
    def run_on_main_thread(self, action):
        wx.CallAfter(action)
    
    def update_ui(self, sender=None):
        self.openGLView.draw()
        label_text = self.controller.player_label_text()
        for i, (score_field,label,popup) in enumerate(zip(*[self.controls[x] for x in ('playerScoreFields', 'playerLabels', 'playerPopups')])):
            # make color 33% to black so light colors will show up
            color = [(2*rgb)//3 for rgb in Preferences.prefs.player_color(i+1)]
            score_field.SetForegroundColour(color)
            label.SetForegroundColour(color)
            score_field.SetLabel(self.controller.score_label(i+1))
            font = label.GetFont()
            font.SetWeight(wx.BOLD if self.controller.current_game_state().current_player==i+1 else wx.NORMAL)
            label.SetFont(font)
            label.SetLabel(label_text[i])
            # disable controls for inactive players (hiding would be better but that causes problems when resizing)
            visible = self.num_players() > i
            score_field.Enable(visible)
            label.Enable(visible)
            popup.Enable(visible)

        # fixme: need more flexible messages
        self.controls['gameStatusField'].SetLabel(self.controller.game_status_message())
        
        self.controls['pass_button'].Enabled = self.controller.can_human_pass()
        self.controls['undo_button'].Enabled = self.controller.can_undo_move()
        self.controls['redo_button'].Enabled = self.controller.can_redo_move()
        
        self.menuitems['undo_menuitem'].Enable(self.controller.can_undo_move())
        self.menuitems['redo_menuitem'].Enable(self.controller.can_redo_move())
        # update the "Game Help" menu item text; this doesn't work on OS X in wxPython 2.8
        self.menuitems['gamehelp_menuitem'].SetText(self.game_config.selected_game_display_name() + ' Help')
            
    def animation_finished(self):
        self.update_ui()
        
    def made_move(self, sender=None, prevstate=None, animations=None):
        self.controller.clear_highlighted_positions()
        if animations and Preferences.prefs.use_animation():
            # If human turn, calculate successors first, so it doesn't interfere with the animation
            # if it takes a long time.
            if self.controller.is_human_turn():
                self.current_game_state().cached_successors()
            self.openGLView.animation_callbacks.append(self.animation_finished)
            self.openGLView.run_animations(animations, self.current_game_state())
        else:
            self.openGLView.game_state = self.current_game_state()
            self.animation_finished()
        
    def do_game_over(self, sender=None):
        # UI gets updated in made_move, nothing special needed here
        pass
                        
    def start_ai_move(self, sender=None):
        # we'd like to let the AI thread start while the animation is running, but that makes it choppy
        # use CallAfter to make sure UI updates correctly
        def go():
            wx.CallAfter(lambda: self.controller.do_ai_move(threaded=True))
            
        if self.openGLView.is_animating():
            self.openGLView.animation_callbacks.append(go)
        else:
            go()
        
    def made_ai_move(self, sender=None, prevstate=None, animations=None):
        # wait for animation to finish, if in progress
        def doit():
            if animations and Preferences.prefs.use_animation(): 
                # use CallAfter in the callback to avoid drawing the "future" state before the next animation
                self.openGLView.animation_callbacks.append(lambda: wx.CallAfter(self.controller.made_move))
                self.openGLView.run_animations(animations, self.current_game_state())
            else:
                self.controller.made_move()
                
        if self.openGLView.is_animating():
            self.openGLView.animation_callbacks.append(doit)
        else:
            doit()
            
    def network_listening(self, sender=None, port=27500, sock=None):
        if self.host_window_controller:
            self.host_window_controller.network_listening(port, sock)
            
    def network_listen_failed(self, sender=None, port=27500, exception=None):
        if self.host_window_controller:
            self.host_window_controller.network_listen_failed(port)

    def network_connected(self, sender=None, isserver=True, sock=None):
        self.updatePlayerPopups()
        if isserver: 
            self.controller.start_game()
        for c in (self.host_window_controller, self.join_window_controller):
            if c: 
                c.network_connected(sock)
                c.hideWindow()
        self.controller.broadcast_player_name()
        self.do_show_chat(None)
        
    def network_game_started(self, sender=None):
        self.updateGameVariationPopups()
        
    def network_chat(self, sender, chat, chat_message):
        if not self.chat_window_controller.is_window_visible():
            self.do_show_chat(None)
        self.chat_window_controller.append_chat_message(chat.player_name, chat_message)
        
    def network_disconnected(self, sender):
        dlg = wx.MessageDialog(self.frame, "The network connection was interrupted.",
                                   "Network Disconnected", wx.ICON_INFORMATION)
        dlg.ShowModal()
        self.updatePlayerPopups()

        
    # game controller delegate methods end

    def highlightNextCell(self, direction=+1):
        self.controller.highlight_next_valid_position(direction)
        
    def highlightCell(self, pos, update=True):
        return self.controller.highlight_position_if_valid(pos)
            
    def undoSelection(self):
        self.controller.undo_partial_move()
        
    def highlightedCells(self):
        return self.controller.partial_move + [self.controller.highlighted_position]
        
    def selectHighlightedCell(self):
        self.controller.select_position(self.controller.highlighted_position)
                
    def doGameChoice(self, event):
        popup = self.controls['gameChoice']
        if self.confirm_new_game():
            game = popup.GetClientData(popup.GetSelection())
            if self.controller.game_config.select_game(game, variation=Preferences.prefs.get_variation_for_game(game)):
                self.updateVariationPopup()
                self.set_default_players()                    
                self.updatePlayerPopups()
                self.start_new_game()
                # store in preferences so this game will start at next launch
                Preferences.prefs.set_initial_game(game)
        else:
            # restore previous selection
            popup.SetSelection(self.game_config.selected_game_index())
        
    def do_variation_choice(self, event):
        popup = self.controls['variationChoice']
        if self.confirm_new_game():
            variation = popup.GetClientData(popup.GetSelection())
            if self.game_config.select_variation(variation):
                Preferences.prefs.set_variation_for_game(self.game_config.game_name, variation)
                self.start_new_game()
        else:
            popup.SetSelection(self.game_config.selected_variation_index())
        
    def updateGameVariationPopups(self):
        self.controls['gameChoice'].SetSelection(self.game_config.selected_game_index())
        self.updateVariationPopup()
        self.controls['variationChoice'].SetSelection(self.game_config.selected_variation_index())
        
    def updateVariationPopup(self):
        self.controls['variationChoice'].Clear()
        for v in self.game_config.variation_items():
            self.controls['variationChoice'].Append(v[0], v[1])
        self.controls['variationChoice'].SetSelection(self.game_config.selected_variation_index())
        
    def doPlayerPopupChoice(self, event):
        popup = event.GetEventObject()
        pnum = 1+self.controls['playerPopups'].index(popup)
        sel = popup.GetClientData(popup.GetSelection())
        if sel==GameConfig.NETWORK_PLAYER:
            # no-op, should only be visible when network game is in progress
            return
        ptype = self.game_config.player_type(pnum)
        cnum = self.game_config.current_player()
        if sel==GameConfig.HUMAN_PLAYER:
            self.game_config.set_human_player(pnum)
            if ptype==GameConfig.NETWORK_PLAYER:
                self.controller.disconnect_network()
            if ptype!=GameConfig.HUMAN_PLAYER and cnum==pnum and not self.controller.is_game_over(): 
                self.controller.made_move()
        else:
            self.game_config.set_ai_for_player(sel, pnum)
            if ptype==GameConfig.NETWORK_PLAYER:
                self.controller.disconnectNetwork()
            if ptype!=GameConfig.AI_PLAYER and cnum==pnum and not self.controller.is_game_over():
                self.start_ai_move()
        # update display fields and store in prefs
        self.update_ui()
        Preferences.prefs.set_ai_for_game(self.game_config.game_name, pnum, sel)
            
        
    def updatePlayerPopup(self, popup, pnum):
        popup.Clear()
        popup.Append("Human", GameConfig.HUMAN_PLAYER)
        for label,value in self.game_config.ai_items():
            popup.Append("AI: "+label, value)
        ptype = self.game_config.player_type(pnum)
        if ptype==GameConfig.HUMAN_PLAYER:
            popup.SetSelection(0)
        elif ptype==GameConfig.AI_PLAYER:
            popup.SetSelection(1+self.game_config.selected_ai_index_for_player(pnum))
        elif ptype==GameConfig.NETWORK_PLAYER:
            popup.Append("Network", GameConfig.NETWORK_PLAYER)
            popup.SetSelection(popup.GetCount()-1)
        
    def updatePlayerPopups(self):
        for i,popup in enumerate(self.controls['playerPopups']):
            self.updatePlayerPopup(popup, i+1)
        
    def do_window_closed(self, event):
        Preferences.prefs.set_window_rect(self.frame.GetRect().Get())
        self.Exit()

    def check_for_update(self, show_all_results=False):
        def no_update():
            if show_all_results:
                dlg = wx.MessageDialog(self.frame, "This version (%s) is the most recent." % (PGAppInfo.version,), 
                                       "Up to date!", wx.OK | wx.ICON_INFORMATION)
                dlg.ShowModal()
                
        def error(exception):
            if show_all_results:
                dlg = wx.MessageDialog(self.frame, "An error occurred checking for updates. Please try again later.", 
                                       "Update check failed", wx.OK | wx.ICON_WARNING)
                dlg.ShowModal()
                
        def update(version, url):
            msg = "Gridlock version %s is available (you have %s). Would you like to download it?" % (version, PGAppInfo.version)
            dlg = wx.MessageDialog(self.frame, msg, "New version available!", wx.YES_NO | wx.ICON_INFORMATION)
            if dlg.ShowModal()==wx.ID_YES:
                import webbrowser
                webbrowser.open_new(url)
            
        after = WXUtils.callafter_wrapper
        PGUtils.check_app_update(PGAppInfo.update_check_url, PGAppInfo.version_number, update_callback=after(update), no_update_callback=after(no_update), error_callback=after(error), threaded=True)

    def do_manual_update_check(self, event):
        self.check_for_update(True)
        

def main():
    app = PGApp(redirect=False)
    app.MainLoop()
    
if __name__=='__main__':
    main()
    